import 'dart:convert';

import 'package:cryptography/cryptography.dart';
import 'package:equatable/equatable.dart';

/// Represents the cryptoBody of a [Ticket] or [Stub]
class CryptoBody extends Equatable {
  CryptoBody.constructor(
      this.hashAlgorithm, this.hash, this.signatureAlgorithm, this.signature);

  factory CryptoBody(HashAlgorithm hashAlgorithm, Hash hash,
      SignatureAlgorithm signatureAlgorithm, Signature signature) {
    if (signatureAlgorithm != ed25519 && signatureAlgorithm != null) {
      throw FormatException('signatureAlgorithm must be `ed25519` or null');
    }

    if (![blake2b, blake2s, sha512, sha256].contains(hashAlgorithm)) {
      throw FormatException(
          'hashAlgorithm must be one of { blake2b, blake2s, sha512, sha256 }');
    }

    if (hash == null) throw FormatException('hash cannot be null');

    if (signatureAlgorithm != null && signature == null) {
      throw FormatException(
          'signature cannot be null if signatureAlgorithm is set');
    }

    if (signatureAlgorithm == null && signature != null) {
      throw FormatException(
          'signatureAlgorithm cannot be null if signature is set');
    }

    return CryptoBody.constructor(
        hashAlgorithm, hash, signatureAlgorithm, signature);
  }

  /// The [HashAlgorithm] to use for checksumming
  final HashAlgorithm hashAlgorithm;

  /// The [Hash] of the [Ticket] or [Stub]'s infoBody stringified
  final Hash hash;

  /// The [SignatureAlgorithm] used to sign a [Ticket] or [Stub]'s infoBody
  ///
  /// Currently only supports [ed25519]
  final SignatureAlgorithm signatureAlgorithm;

  /// The [Signature] of a [Ticket] or [Stub]'s infoBody
  final Signature signature;

  /// Returns a [bool] representing whether or not the [hash] matches given data. (Synchronous)
  bool checksumMatchesSync(List<int> data) {
    return hashAlgorithm.hashSync(data) == hash;
  }

  /// Returns a [bool] representing whether or not the [hash] matches given data.
  Future<bool> checksumMatches(List<int> data) async {
    return await hashAlgorithm.hash(data) == hash;
  }

  /// Checks if the [signature] matches a [PublicKey] and given data. (Synchronous)
  bool signatureMatchesSync(PublicKey publicKey, List<int> data) {
    if (signature == null) {
      throw FormatException(
          'cannot check if the signature matches for a ticket without a signature');
    }
    var isSignatureValid = signatureAlgorithm.verifySync(data, signature);
    var publicKeyMatches = publicKey == signature.publicKey;
    if (isSignatureValid && publicKeyMatches) {
      return true;
    } else {
      return false;
    }
  }

  /// Checks if the [signature] matches a [PublicKey] and given data.
  Future<bool> signatureMatches(PublicKey publicKey, List<int> data) async {
    if (signature == null) {
      throw FormatException(
          'cannot check if the signature matches for a ticket without a signature');
    }
    var isSignatureValid = signatureAlgorithm.verify(data, signature);
    var publicKeyMatches = publicKey == signature.publicKey;
    if (await isSignatureValid && publicKeyMatches) {
      return true;
    } else {
      return false;
    }
  }

  /// Checks if this [CryptoBody] is valid for given data.
  ///
  /// If the [publicKey] is null, signature checking will be skipped.
  Future<bool> isValid(List<int> data, PublicKey publicKey) async {
    bool signatureOkay;

    if (publicKey == null) {
      signatureOkay = true;
    } else {
      signatureOkay = await signatureMatches(publicKey, data);
    }

    return (await checksumMatches(data) && signatureOkay);
  }

  /// Checks if this [CryptoBody] is valid for given data. (Synchronous)
  ///
  /// If the [publicKey] is null, signature checking will be skipped.
  bool isValidSync(List<int> data, PublicKey publicKey) {
    bool signatureOkay;

    if (publicKey == null) {
      signatureOkay = true;
    } else {
      signatureOkay = signatureMatchesSync(publicKey, data);
    }

    return (checksumMatchesSync(data) && signatureOkay);
  }

  /// Returns a [Map] representing this [CryptoBody].
  Map toMap() {
    String sigAlgoName;
    List<int> sigBytes;
    List<int> pubKeyBytes;

    if (signatureAlgorithm == null) {
      sigAlgoName = null;
      sigBytes = null;
      pubKeyBytes = null;
    } else {
      sigAlgoName = signatureAlgorithm.name;
      sigBytes = signature.bytes;
      pubKeyBytes = signature.publicKey.bytes;
    }

    return {
      'hashAlgorithm': hashAlgorithm.name,
      'hash': hash.bytes,
      'signatureAlgorithm': sigAlgoName,
      'signatureBytes': sigBytes,
      'signaturePublicKeyBytes': pubKeyBytes
    };
  }

  /// Returns a [CryptoBody], reconstructed from a given [Map]
  factory CryptoBody.fromMap(Map map) {
    HashAlgorithm hashAlgo;

    switch (map['hashAlgorithm']) {
      case 'blake2b':
        {
          hashAlgo = blake2b;
          break;
        }
      case 'blake2s':
        {
          hashAlgo = blake2s;
          break;
        }
      case 'sha512':
        {
          hashAlgo = sha512;
          break;
        }
      case 'sha256':
        {
          hashAlgo = sha256;
          break;
        }
    }

    if (map['signatureAlgorithm'] == null) {
      return CryptoBody(
          hashAlgo, Hash(List<int>.from(map['hash'])), null, null);
    } else {
      return CryptoBody(
          hashAlgo,
          Hash(List<int>.from(map['hash'])),
          ed25519,
          Signature(List<int>.from(map['signatureBytes']),
              publicKey:
                  PublicKey(List<int>.from(map['signaturePublicKeyBytes']))));
    }
  }

  /// Returns a JSON string representing this [CryptoBody]
  String toJson() {
    return jsonEncode(toMap());
  }

  /// Returns a [CryptoBody] from a JSON string
  factory CryptoBody.fromJson(String json) {
    return CryptoBody.fromMap(jsonDecode(json));
  }

  @override
  List<Object> get props =>
      [hashAlgorithm, hash, signatureAlgorithm, signature];
}
