import 'dart:convert';

import 'package:cryptography/cryptography.dart';
import 'package:equatable/equatable.dart';
import 'package:meta/meta.dart';
import 'package:ticket/ticket.dart';
import 'package:version/version.dart';

import '../ticket.dart';

/// Represents a Stub.
///
/// Implements validity checking, converting to and from maps, converting to and from JSON.
/// Includes Generators.
class Stub extends Equatable {
  Stub(this.cryptoBody, this.infoBody, this.specVersion);

  /// The [CryptoBody] of the [Ticket] instance
  final CryptoBody cryptoBody;

  /// The [StubInfoBody] of the [Ticket] instance
  final StubInfoBody infoBody;

  /// The version of the specification that the [Stub] instance adheres to
  final Version specVersion;

  /// Checks if the [Stub] instance is valid.
  Future<bool> isValid() async {
    if (infoBody.isValid() == false) {
      return false;
    }

    if (await cryptoBody.isValid(utf8.encode(infoBody.toJson()), null) ==
        false) {
      return false;
    }

    return true;
  }

  /// Checks if the [Stub] instance is valid. (Synchronously)
  bool isValidSync() {
    if (infoBody.isValid() == false) {
      return false;
    }

    if (cryptoBody.isValidSync(utf8.encode(infoBody.toJson()), null) == false) {
      return false;
    }

    return true;
  }

  /// Returns a [Map] representing the [Stub] instance.
  Map toMap() {
    return {
      'cryptoBody': cryptoBody.toMap(),
      'infoBody': infoBody.toMap(),
      'specVersion': specVersion.toString()
    };
  }

  /// Returns stringified JSON representing the [Stub] instance.
  String toJson() {
    return jsonEncode(toMap());
  }

  /// Returns a [Stub], reconstructed from a [Map] generated by [toMap()].
  factory Stub.fromMap(Map map) {
    var cryptoBody = CryptoBody.fromMap(map['cryptoBody']);
    var infoBody = StubInfoBody.fromMap(map['infoBody']);
    var version = Version.parse(map['specVersion']);
    return Stub(cryptoBody, infoBody, version);
  }

  /// Returns a [Stub], reconstructed from stringified JSON.
  factory Stub.fromJson(String json) {
    return Stub.fromMap(jsonDecode(json));
  }

  /// Returns a [Stub], generated using the given information.
  ///
  /// If [keyPair] is set to something other than null, the generated [Stub] will be signed.
  factory Stub.generate(
      {@required List<int> transactionIdentifier,
      dynamic payload,
      bool error,
      HashAlgorithm hashAlgorithm = blake2b,
      @required KeyPair keyPair}) {
    var infoBody = StubInfoBody(transactionIdentifier, payload, error);

    // break the infoBody into bytes here as we may be using it as crypto input in multiple places, depending on if we're signing the ticket
    var infoBodyBytes = utf8.encode(infoBody.toJson());

    // initialize these to null as we might want to pass them as either null or with content
    SignatureAlgorithm signatureAlgorithm;
    Signature signature;

    if (keyPair != null) {
      signatureAlgorithm = ed25519;
      signature = signatureAlgorithm.signSync(infoBodyBytes, keyPair);
    }

    var cryptoBody = CryptoBody(hashAlgorithm,
        hashAlgorithm.hashSync(infoBodyBytes), signatureAlgorithm, signature);

    return Stub(cryptoBody, infoBody, Version(4, 0, 0));
  }

  @override
  List<Object> get props => [cryptoBody, infoBody, specVersion];
}
