import 'package:cryptography/cryptography.dart';
import 'package:equatable/equatable.dart';
import 'package:meta/meta.dart';
import 'package:ticket/ticket.dart';

/// Utility class for generating many [Ticket]s using the same [KeyPair], [serviceName], and other options
///
/// [validityPeriod] defaults to 30 seconds
///
/// [hashAlgorithm] defaults to [blake2b]
///
/// Instantiate a [TGen] instance, and call [generate()] on that to generate [Ticket]s while only needing to provide an operation and payload, the rest is filled in by [TGen]
class TGen extends Equatable {
  TGen._constructor(
      this.serviceName, this.keyPair, this.validityPeriod, this.hashAlgorithm);

  factory TGen(
      {@required String serviceName,
      @required KeyPair keyPair,
      Duration validityPeriod,
      HashAlgorithm hashAlgorithm = blake2b}) {
    validityPeriod ??= Duration(seconds: 30000);

    return TGen._constructor(
        serviceName, keyPair, validityPeriod, hashAlgorithm);
  }

  final String serviceName;
  final KeyPair keyPair;
  final Duration validityPeriod;
  final HashAlgorithm hashAlgorithm;

  /// Generate a [Ticket] using the [KeyPair] and other options provided when instantiating [TGen]
  Ticket generate({@required String operation, @required dynamic payload}) {
    return Ticket.generate(
        serviceName: serviceName,
        operation: operation,
        keyPair: keyPair,
        validityPeriod: validityPeriod,
        hashAlgorithm: hashAlgorithm);
  }

  @override
  List<Object> get props =>
      [serviceName, keyPair, validityPeriod, hashAlgorithm];
}
