import 'dart:convert';

import 'package:cryptography/cryptography.dart';
import 'package:equatable/equatable.dart';
import 'package:meta/meta.dart';
import 'package:ticket/ticket.dart';
import 'package:version/version.dart';

import '../ticket.dart';

/// Represents a Ticket.
///
/// Implements validity checking, converting to and from maps, converting to and from JSON.
/// Includes Generators.
class Ticket extends Equatable {
  Ticket(this.cryptoBody, this.infoBody, this.specVersion);

  /// The [CryptoBody] of the [Ticket] instance
  final CryptoBody cryptoBody;

  /// The [InfoBody] of the [Ticket] instance
  final InfoBody infoBody;

  /// The version of the specification that the [Ticket] instance adheres to
  final Version specVersion;

  /// Checks if the [Ticket] instance is valid.
  Future<bool> isValid() async {
    if (infoBody.isValid() == false) {
      return false;
    }

    if (await cryptoBody.isValid(utf8.encode(infoBody.toJson()), null) ==
        false) {
      return false;
    }

    return true;
  }

  /// Checks if the [Ticket] instance is valid. (Synchronously)
  bool isValidSync() {
    if (infoBody.isValid() == false) {
      return false;
    }

    if (cryptoBody.isValidSync(utf8.encode(infoBody.toJson()), null) == false) {
      return false;
    }

    return true;
  }

  /// Returns a [Map] representing the [Ticket] instance.
  Map toMap() {
    return {
      'cryptoBody': cryptoBody.toMap(),
      'infoBody': infoBody.toMap(),
      'specVersion': specVersion.toString()
    };
  }

  /// Returns stringified JSON representing the [Ticket] instance.
  String toJson() {
    return jsonEncode(toMap());
  }

  /// Returns a [Ticket], reconstructed from a [Map] generated by [toMap()].
  factory Ticket.fromMap(Map map) {
    var cryptoBody = CryptoBody.fromMap(map['cryptoBody']);
    var infoBody = InfoBody.fromMap(map['infoBody']);
    var version = Version.parse(map['specVersion']);
    return Ticket(cryptoBody, infoBody, version);
  }

  /// Returns a [Ticket], reconstructed from stringified JSON.
  factory Ticket.fromJson(String json) {
    return Ticket.fromMap(jsonDecode(json));
  }

  /// Returns a [Ticket], generated using the given information.
  ///
  /// [payload] should be any serializable value. (e.g. able to be converted to a map then to a JSON string)
  ///
  /// [serviceName] and [operation] should be the name of the service and operation you're trying to use
  ///
  /// [timestamp] controls the timestamp field in the ticket's [InfoBody] - nullable, you probably shouldn't change this - defaults to the current timestamp
  ///
  /// [transactionIdentifier] controls the transaction identifier - you probably shouldn't change this - defaults to random 12-byte nonce
  ///
  /// [validityPeriod] controls how long the ticket is valid for - default is 30 seconds
  ///
  /// [hashAlgorithm] sets the hashing algorithm used when generating the [cryptoBody]
  ///
  /// If [keyPair] is left `null`, the ticket will not be signed. If [keyPair] is non-null, the ticket will be signed.
  factory Ticket.generate(
      {dynamic payload,
      @required String serviceName,
      @required String operation,
      int timestamp,
      Duration validityPeriod,
      List<int> transactionIdentifier,
      HashAlgorithm hashAlgorithm = blake2b,
      @required KeyPair keyPair}) {
    // if there isn't a specified timestamp, use the current one
    timestamp ??= DateTime.now().millisecondsSinceEpoch;

    // set validityPeriod to 30 seconds if not specified
    validityPeriod ??= Duration(seconds: 30);

    // generate transactionIdentifier if one isn't specified
    transactionIdentifier ??= Nonce.randomBytes(12).bytes;

    // generate infoBody
    var infoBody = InfoBody(serviceName, operation, transactionIdentifier,
        timestamp, validityPeriod, payload);

    // break the infoBody into bytes here as we may be using it as crypto input in multiple places, depending on if we're signing the ticket
    var infoBodyBytes = utf8.encode(infoBody.toJson());

    // initialize these to null as we might want to pass them as either null or with content
    SignatureAlgorithm signatureAlgorithm;
    Signature signature;

    if (keyPair != null) {
      signatureAlgorithm = ed25519;
      signature = signatureAlgorithm.signSync(infoBodyBytes, keyPair);
    }

    var cryptoBody = CryptoBody(hashAlgorithm,
        hashAlgorithm.hashSync(infoBodyBytes), signatureAlgorithm, signature);

    return Ticket(cryptoBody, infoBody, Version(4, 0, 0));
  }

  @override
  List<Object> get props => [cryptoBody, infoBody, specVersion];
}
