import 'dart:convert';

import 'package:cryptography/cryptography.dart';
import 'package:ticket/ticket.dart';
import 'package:test/test.dart';

void main() {
  group('CryptoBody', () {
    group('Initialization', () {
      test('throws FormatException on bad hashAlgorithm', () {
        expect(() {
          CryptoBody(sha224, Hash([1]), null, null);
        }, throwsA(isA<FormatException>()));

        expect(() {
          CryptoBody(null, Hash([1]), null, null);
        }, throwsA(isA<FormatException>()));
      });

      test('throws FormatException on null hash', () {
        expect(() {
          CryptoBody(blake2b, null, null, null);
        }, throwsA(isA<FormatException>()));
      });

      test('throws FormatException on bad signatureAlgorithm', () {
        expect(() {
          CryptoBody(blake2b, Hash([1]), ecdsaP521Sha256, null);
        }, throwsA(isA<FormatException>()));
      });

      test(
          'throws FormatException on null signature when signatureAlgorithm is set',
          () {
        expect(() {
          CryptoBody(blake2b, Hash([1]), ed25519, null);
        }, throwsA(isA<FormatException>()));
      });

      test(
          'throws FormatException on null signatureAlgorithm when signature is set',
          () {
        expect(() {
          CryptoBody(blake2b, Hash([1]), null,
              Signature([1], publicKey: PublicKey([1])));
        }, throwsA(isA<FormatException>()));
      });
    });

    [blake2b, blake2s, sha256, sha512].forEach((element) {
      group(element.name, () {
        test('toMap()', () {
          expect(
              CryptoBody(element, Hash([1]), ed25519,
                  Signature([1], publicKey: PublicKey([1]))).toMap(),
              equals({
                'hashAlgorithm': element.name,
                'hash': [1],
                'signatureAlgorithm': ed25519.name,
                'signatureBytes': [1],
                'signaturePublicKeyBytes': [1]
              }));

          expect(
              CryptoBody(element, Hash([1]), null, null).toMap(),
              equals({
                'hashAlgorithm': element.name,
                'hash': [1],
                'signatureAlgorithm': null,
                'signatureBytes': null,
                'signaturePublicKeyBytes': null
              }));
        });

        test('toJson()', () {
          expect(
              CryptoBody(element, Hash([1]), ed25519,
                  Signature([1], publicKey: PublicKey([1]))).toJson(),
              equals(jsonEncode({
                'hashAlgorithm': element.name,
                'hash': [1],
                'signatureAlgorithm': ed25519.name,
                'signatureBytes': [1],
                'signaturePublicKeyBytes': [1]
              })));
        });

        test('fromMap()', () {
          expect(
              CryptoBody.fromMap(CryptoBody(element, Hash([1]), ed25519,
                  Signature([1], publicKey: PublicKey([1]))).toMap()),
              equals(CryptoBody(element, Hash([1]), ed25519,
                  Signature([1], publicKey: PublicKey([1])))));
        });

        test('fromJson()', () {
          expect(
              CryptoBody.fromJson(CryptoBody(element, Hash([1]), ed25519,
                  Signature([1], publicKey: PublicKey([1]))).toJson()),
              equals(CryptoBody(element, Hash([1]), ed25519,
                  Signature([1], publicKey: PublicKey([1])))));
        });

        group('checksumMatchesSync()', () {
          CryptoBody goodBody;
          CryptoBody badBody;

          setUp(() {
            goodBody = CryptoBody(
                element, element.hashSync(<int>[0, 1, 2]), null, null);
            badBody = CryptoBody(element, Hash(<int>[1]), null, null);
          });

          test('returns true for good checksum', () {
            expect(goodBody.checksumMatchesSync([0, 1, 2]), equals(true));
          });

          test('returns false for bad checksum', () {
            expect(badBody.checksumMatchesSync([0, 1, 2]), equals(false));
          });
        });

        group('checksumMatches()', () {
          CryptoBody goodBody;
          CryptoBody badBody;

          setUp(() {
            goodBody = CryptoBody(
                element, element.hashSync(<int>[0, 1, 2]), null, null);
            badBody = CryptoBody(element, Hash(<int>[1]), null, null);
          });

          test('returns true for good checksum', () async {
            expect(await goodBody.checksumMatches([0, 1, 2]), equals(true));
          });

          test('returns false for bad checksum', () async {
            expect(await badBody.checksumMatches([0, 1, 2]), equals(false));
          });
        });
      });
    });

    group('signatureMatchesSync()', () {
      test('throws FormatException on null signature', () {
        expect(() {
          CryptoBody(blake2b, Hash([1]), null, null).signatureMatchesSync(
              ed25519.newKeyPairSync().publicKey, <int>[0]);
        }, throwsA(isA<FormatException>()));
      });

      test('returns true when everything is good', () {
        var data = <int>[0, 1, 2];
        var keyPair = ed25519.newKeyPairSync();

        var cryptoBody = CryptoBody(blake2b, blake2b.hashSync(data), ed25519,
            ed25519.signSync(data, keyPair));

        expect(cryptoBody.signatureMatchesSync(keyPair.publicKey, data),
            equals(true));
      });

      test('returns false when the publicKey does not match', () {
        var data = <int>[0, 1, 2];
        var badPublicKey = ed25519.newKeyPairSync().publicKey;

        var goodCryptoBody = CryptoBody(blake2b, blake2b.hashSync(data),
            ed25519, ed25519.signSync(data, ed25519.newKeyPairSync()));

        expect(goodCryptoBody.signatureMatchesSync(badPublicKey, data),
            equals(false));
      });

      test('returns false when the data does not match', () {
        var data = <int>[0, 1, 2];
        var keyPair = ed25519.newKeyPairSync();

        var cryptoBody = CryptoBody(blake2b, blake2b.hashSync(data), ed25519,
            ed25519.signSync(data, keyPair));

        expect(
            cryptoBody.signatureMatchesSync(keyPair.publicKey, <int>[3, 4, 5]),
            equals(false));
      });

      test('returns false when nothing is correct', () {
        var data = <int>[0, 1, 2];
        var keyPair = ed25519.newKeyPairSync();

        var cryptoBody = CryptoBody(blake2b, blake2b.hashSync(data), ed25519,
            ed25519.signSync(data, keyPair));

        expect(
            cryptoBody.signatureMatchesSync(
                ed25519.newKeyPairSync().publicKey, <int>[3, 4, 5]),
            equals(false));
      });
    });

    group('signatureMatches()', () {
      test('throws FormatException on null signature', () {
        expect(() async {
          await CryptoBody(blake2b, Hash([1]), null, null)
              .signatureMatches(ed25519.newKeyPairSync().publicKey, <int>[0]);
        }, throwsA(isA<FormatException>()));
      });

      test('returns true when everything is good', () async {
        var data = <int>[0, 1, 2];
        var keyPair = await ed25519.newKeyPair();

        var cryptoBody = CryptoBody(blake2b, blake2b.hashSync(data), ed25519,
            ed25519.signSync(data, keyPair));

        expect(await cryptoBody.signatureMatches(keyPair.publicKey, data),
            equals(true));
      });

      test('returns false when the publicKey does not match', () async {
        var data = <int>[0, 1, 2];
        var badPublicKey = ed25519.newKeyPairSync().publicKey;

        var goodCryptoBody = CryptoBody(blake2b, blake2b.hashSync(data),
            ed25519, ed25519.signSync(data, ed25519.newKeyPairSync()));

        expect(await goodCryptoBody.signatureMatches(badPublicKey, data),
            equals(false));
      });

      test('returns false when the data does not match', () async {
        var data = <int>[0, 1, 2];
        var keyPair = ed25519.newKeyPairSync();

        var cryptoBody = CryptoBody(blake2b, blake2b.hashSync(data), ed25519,
            ed25519.signSync(data, keyPair));

        expect(
            await cryptoBody
                .signatureMatches(keyPair.publicKey, <int>[3, 4, 5]),
            equals(false));
      });

      test('returns false when nothing is correct', () async {
        var data = <int>[0, 1, 2];
        var keyPair = ed25519.newKeyPairSync();

        var cryptoBody = CryptoBody(blake2b, blake2b.hashSync(data), ed25519,
            ed25519.signSync(data, keyPair));

        expect(
            await cryptoBody.signatureMatches(
                ed25519.newKeyPairSync().publicKey, <int>[3, 4, 5]),
            equals(false));
      });
    });

    group('isValid()', () {
      group('returns true when everything is fine', () {
        test('with a signature', () async {
          var data = <int>[0, 1, 2];
          var keypair = await ed25519.newKeyPair();

          var cryptoBody = CryptoBody(blake2b, await blake2b.hash(data),
              ed25519, await ed25519.sign(data, keypair));

          expect(
              await cryptoBody.isValid(data, keypair.publicKey), equals(true));
        });

        test('without a signature', () async {
          var data = <int>[0, 1, 2];

          var cryptoBody =
              CryptoBody(blake2b, await blake2b.hash(data), null, null);

          expect(await cryptoBody.isValid(data, null), equals(true));
        });
      });

      group('returns false', () {
        test("when the checksum doesn't match", () async {
          var data = <int>[0, 1, 2];

          var cryptoBody = CryptoBody(
              blake2b, await blake2b.hash(<int>[5, 3, 6]), null, null);

          expect(await cryptoBody.isValid(data, null), equals(false));
        });

        test("when the signature doesn't match", () async {
          var data = <int>[0, 1, 2];
          var goodKeypair = await ed25519.newKeyPair();
          var badKeypair = await ed25519.newKeyPair();

          var cryptoBody = CryptoBody(blake2b, await blake2b.hash(data),
              ed25519, await ed25519.sign(data, goodKeypair));

          expect(await cryptoBody.isValid(data, badKeypair.publicKey),
              equals(false));
        });

        test('when nothing matches', () async {
          var data = <int>[0, 1, 2];
          var goodKeypair = await ed25519.newKeyPair();
          var badKeypair = await ed25519.newKeyPair();

          var cryptoBody = CryptoBody(
              blake2b,
              await blake2b.hash(<int>[5, 3, 6]),
              ed25519,
              await ed25519.sign(data, goodKeypair));

          expect(await cryptoBody.isValid(data, badKeypair.publicKey),
              equals(false));
        });
      });
    });

    group('isValidSync()', () {
      group('returns true when everything is fine', () {
        test('with a signature', () {
          var data = <int>[0, 1, 2];
          var keypair = ed25519.newKeyPairSync();

          var cryptoBody = CryptoBody(blake2b, blake2b.hashSync(data), ed25519,
              ed25519.signSync(data, keypair));

          expect(cryptoBody.isValidSync(data, keypair.publicKey), equals(true));
        });

        test('without a signature', () {
          var data = <int>[0, 1, 2];

          var cryptoBody =
              CryptoBody(blake2b, blake2b.hashSync(data), null, null);

          expect(cryptoBody.isValidSync(data, null), equals(true));
        });
      });

      group('returns false', () {
        test("when the checksum doesn't match", () {
          var data = <int>[0, 1, 2];

          var cryptoBody =
              CryptoBody(blake2b, blake2b.hashSync(<int>[5, 3, 6]), null, null);

          expect(cryptoBody.isValidSync(data, null), equals(false));
        });

        test("when the signature doesn't match", () {
          var data = <int>[0, 1, 2];
          var goodKeypair = ed25519.newKeyPairSync();
          var badKeypair = ed25519.newKeyPairSync();

          var cryptoBody = CryptoBody(blake2b, blake2b.hashSync(data), ed25519,
              ed25519.signSync(data, goodKeypair));

          expect(cryptoBody.isValidSync(data, badKeypair.publicKey),
              equals(false));
        });

        test('when nothing matches', () {
          var data = <int>[0, 1, 2];
          var goodKeypair = ed25519.newKeyPairSync();
          var badKeypair = ed25519.newKeyPairSync();

          var cryptoBody = CryptoBody(blake2b, blake2b.hashSync(<int>[5, 3, 6]),
              ed25519, ed25519.signSync(data, goodKeypair));

          expect(cryptoBody.isValidSync(data, badKeypair.publicKey),
              equals(false));
        });
      });
    });
  });
}
