import 'dart:convert';

import 'package:cryptography/cryptography.dart';
import 'package:ticket/ticket.dart';
import 'package:test/test.dart';
import 'package:version/version.dart';

void main() {
  group('Stub', () {
    Stub unsignedTestStub;
    Stub signedTestStub;
    CryptoBody unsignedCryptoBody;
    CryptoBody signedCryptoBody;
    KeyPair keyPair;
    StubInfoBody infoBody;

    setUp(() {
      infoBody = StubInfoBody([0, 0, 0, 0, 0, 0], null, false);

      unsignedCryptoBody = CryptoBody(blake2b,
          blake2b.hashSync(utf8.encode(infoBody.toJson())), null, null);

      keyPair = ed25519.newKeyPairSync();

      signedCryptoBody = CryptoBody(
          blake2b,
          blake2b.hashSync(utf8.encode(infoBody.toJson())),
          ed25519,
          ed25519.signSync(utf8.encode(infoBody.toJson()), keyPair));

      unsignedTestStub = Stub(unsignedCryptoBody, infoBody, Version(4, 0, 0));
      signedTestStub = Stub(signedCryptoBody, infoBody, Version(4, 0, 0));
    });

    test('toMap()', () {
      expect(
          unsignedTestStub.toMap(),
          equals({
            'cryptoBody': unsignedCryptoBody.toMap(),
            'infoBody': infoBody.toMap(),
            'specVersion': unsignedTestStub.specVersion.toString()
          }));

      expect(
          signedTestStub.toMap(),
          equals({
            'cryptoBody': signedCryptoBody.toMap(),
            'infoBody': infoBody.toMap(),
            'specVersion': signedTestStub.specVersion.toString()
          }));
    });

    test('fromMap()', () {
      expect(
          Stub.fromMap({
            'cryptoBody': unsignedCryptoBody.toMap(),
            'infoBody': infoBody.toMap(),
            'specVersion': unsignedTestStub.specVersion.toString()
          }),
          equals(unsignedTestStub));

      expect(
          Stub.fromMap({
            'cryptoBody': signedCryptoBody.toMap(),
            'infoBody': infoBody.toMap(),
            'specVersion': signedTestStub.specVersion.toString()
          }),
          equals(signedTestStub));
    });

    test('toJson()', () {
      expect(unsignedTestStub.toJson(),
          equals(jsonEncode(unsignedTestStub.toMap())));

      expect(
          signedTestStub.toJson(), equals(jsonEncode(signedTestStub.toMap())));
    });

    test('fromJson()', () {
      expect(
          Stub.fromJson(unsignedTestStub.toJson()), equals(unsignedTestStub));

      expect(Stub.fromJson(signedTestStub.toJson()), equals(signedTestStub));
    });

    test('isValid()', () async {
      expect(await unsignedTestStub.isValid(), equals(true));
      expect(await signedTestStub.isValid(), equals(true));
      expect(
          await Stub(
                  CryptoBody(blake2b, Hash([1]), ed25519,
                      Signature([1], publicKey: PublicKey([1]))),
                  StubInfoBody([0, 0, 0, 0, 0, 0], null, null),
                  Version(0, 0, 1))
              .isValid(),
          equals(false));
    });

    test('isValidSync()', () {
      expect(unsignedTestStub.isValidSync(), equals(true));
      expect(signedTestStub.isValidSync(), equals(true));
      expect(
          Stub(
                  CryptoBody(blake2b, Hash([1]), ed25519,
                      Signature([1], publicKey: PublicKey([1]))),
                  StubInfoBody([0, 0, 0, 0, 0, 0], null, null),
                  Version(0, 0, 1))
              .isValidSync(),
          equals(false));
    });

    group('generate()', () {
      List<int> transactionIdentifier;
      dynamic payload;
      bool error;
      HashAlgorithm hashAlgorithm;
      KeyPair keyPair;

      setUp(() {
        payload = 'h';
        hashAlgorithm = blake2b;
        keyPair = ed25519.newKeyPairSync();
        error = false;
        transactionIdentifier = Nonce.randomBytes(12).bytes;
      });

      test('unsigned', () {
        var infoBody = StubInfoBody(transactionIdentifier, payload, error);

        var cryptoBody = CryptoBody(hashAlgorithm,
            hashAlgorithm.hashSync(utf8.encode(infoBody.toJson())), null, null);

        var manualStub = Stub(cryptoBody, infoBody, Version(4, 0, 0));

        var generatedStub = Stub.generate(
            keyPair: null,
            payload: payload,
            transactionIdentifier: transactionIdentifier,
            hashAlgorithm: hashAlgorithm,
            error: false);

        expect(generatedStub, equals(manualStub));
      });

      test('signed', () {
        var infoBody = StubInfoBody(transactionIdentifier, payload, error);

        var cryptoBody = CryptoBody(
            hashAlgorithm,
            hashAlgorithm.hashSync(utf8.encode(infoBody.toJson())),
            ed25519,
            ed25519.signSync(utf8.encode(infoBody.toJson()), keyPair));

        var manualStub = Stub(cryptoBody, infoBody, Version(4, 0, 0));

        var generatedStub = Stub.generate(
            keyPair: keyPair,
            payload: payload,
            transactionIdentifier: transactionIdentifier,
            hashAlgorithm: hashAlgorithm,
            error: false);

        expect(generatedStub, equals(manualStub));
      });
    });
  });
}
