import 'dart:convert';

import 'package:cryptography/cryptography.dart';
import 'package:ticket/ticket.dart';
import 'package:test/test.dart';
import 'package:version/version.dart';

void main() {
  group('Ticket', () {
    Ticket unsignedTestTicket;
    Ticket signedTestTicket;
    CryptoBody unsignedCryptoBody;
    CryptoBody signedCryptoBody;
    KeyPair keyPair;
    InfoBody infoBody;

    setUp(() {
      infoBody = InfoBody('serviceName', 'operation', [0, 0, 0, 0, 0, 0],
          DateTime.now().millisecondsSinceEpoch, Duration(seconds: 60), 'h');

      unsignedCryptoBody = CryptoBody(blake2b,
          blake2b.hashSync(utf8.encode(infoBody.toJson())), null, null);

      keyPair = ed25519.newKeyPairSync();

      signedCryptoBody = CryptoBody(
          blake2b,
          blake2b.hashSync(utf8.encode(infoBody.toJson())),
          ed25519,
          ed25519.signSync(utf8.encode(infoBody.toJson()), keyPair));

      unsignedTestTicket =
          Ticket(unsignedCryptoBody, infoBody, Version(4, 0, 0));

      signedTestTicket = Ticket(signedCryptoBody, infoBody, Version(4, 0, 0));
    });

    test('toMap()', () {
      expect(
          unsignedTestTicket.toMap(),
          equals({
            'cryptoBody': unsignedCryptoBody.toMap(),
            'infoBody': infoBody.toMap(),
            'specVersion': unsignedTestTicket.specVersion.toString()
          }));

      expect(
          signedTestTicket.toMap(),
          equals({
            'cryptoBody': signedCryptoBody.toMap(),
            'infoBody': infoBody.toMap(),
            'specVersion': signedTestTicket.specVersion.toString()
          }));
    });

    test('fromMap()', () {
      expect(
          Ticket.fromMap({
            'cryptoBody': unsignedCryptoBody.toMap(),
            'infoBody': infoBody.toMap(),
            'specVersion': unsignedTestTicket.specVersion.toString()
          }),
          equals(unsignedTestTicket));

      expect(
          Ticket.fromMap({
            'cryptoBody': signedCryptoBody.toMap(),
            'infoBody': infoBody.toMap(),
            'specVersion': signedTestTicket.specVersion.toString()
          }),
          equals(signedTestTicket));
    });

    test('toJson()', () {
      expect(unsignedTestTicket.toJson(),
          equals(jsonEncode(unsignedTestTicket.toMap())));

      expect(signedTestTicket.toJson(),
          equals(jsonEncode(signedTestTicket.toMap())));
    });

    test('fromJson()', () {
      expect(Ticket.fromJson(unsignedTestTicket.toJson()),
          equals(unsignedTestTicket));

      expect(
          Ticket.fromJson(signedTestTicket.toJson()), equals(signedTestTicket));
    });

    test('isValid()', () async {
      expect(await unsignedTestTicket.isValid(), equals(true));
      expect(await signedTestTicket.isValid(), equals(true));
      expect(
          await Ticket(
                  CryptoBody(blake2b, Hash([1]), ed25519,
                      Signature([1], publicKey: PublicKey([1]))),
                  InfoBody(
                      'serviceName',
                      'operation',
                      [0, 0, 0, 0, 0],
                      DateTime.now().millisecondsSinceEpoch,
                      Duration(days: -943898),
                      'h'),
                  Version(0, 0, 1))
              .isValid(),
          equals(false));
    });

    test('isValidSync()', () {
      expect(unsignedTestTicket.isValidSync(), equals(true));
      expect(signedTestTicket.isValidSync(), equals(true));
      expect(
          Ticket(
                  CryptoBody(blake2b, Hash([1]), ed25519,
                      Signature([1], publicKey: PublicKey([1]))),
                  InfoBody(
                      'serviceName',
                      'operation',
                      [0, 0, 0, 0, 0],
                      DateTime.now().millisecondsSinceEpoch,
                      Duration(days: -10),
                      'h'),
                  Version(0, 0, 1))
              .isValidSync(),
          equals(false));
    });

    group('generate()', () {
      int timestamp;
      Duration validityPeriod;
      dynamic payload;
      String serviceName;
      String operation;
      List<int> transactionIdentifier;
      HashAlgorithm hashAlgorithm;
      KeyPair keyPair;

      setUp(() {
        timestamp = DateTime.now().millisecondsSinceEpoch;
        payload = 'h';
        serviceName = 'testService';
        operation = 'test';
        hashAlgorithm = blake2b;
        keyPair = ed25519.newKeyPairSync();
        validityPeriod = Duration(seconds: 60);
        transactionIdentifier = Nonce.randomBytes(12).bytes;
      });

      test('unsigned', () {
        var infoBody = InfoBody(serviceName, operation, transactionIdentifier,
            timestamp, validityPeriod, payload);

        var cryptoBody = CryptoBody(hashAlgorithm,
            hashAlgorithm.hashSync(utf8.encode(infoBody.toJson())), null, null);

        var manualTicket = Ticket(cryptoBody, infoBody, Version(4, 0, 0));

        var generatedTicket = Ticket.generate(
            serviceName: serviceName,
            operation: operation,
            keyPair: null,
            timestamp: timestamp,
            validityPeriod: validityPeriod,
            payload: payload,
            transactionIdentifier: transactionIdentifier,
            hashAlgorithm: hashAlgorithm);

        expect(generatedTicket, equals(manualTicket));
      });

      test('signed', () {
        var infoBody = InfoBody(serviceName, operation, transactionIdentifier,
            timestamp, validityPeriod, payload);

        var cryptoBody = CryptoBody(
            hashAlgorithm,
            hashAlgorithm.hashSync(utf8.encode(infoBody.toJson())),
            ed25519,
            ed25519.signSync(utf8.encode(infoBody.toJson()), keyPair));

        var manualTicket = Ticket(cryptoBody, infoBody, Version(4, 0, 0));

        var generatedTicket = Ticket.generate(
            serviceName: serviceName,
            operation: operation,
            keyPair: keyPair,
            timestamp: timestamp,
            validityPeriod: validityPeriod,
            payload: payload,
            transactionIdentifier: transactionIdentifier,
            hashAlgorithm: hashAlgorithm);

        expect(generatedTicket, equals(manualTicket));
      });
    });
  });
}
